from __future__ import annotations

from pathlib import Path

from tempfile import TemporaryDirectory

from typing import Optional, Union, List, Tuple, Dict

from PySide2 import QtCore

from sohstationviewer.controller.util import display_tracking_info
from sohstationviewer.conf import constants
from sohstationviewer.view.plotting.gps_plot.gps_point import GPSPoint
from sohstationviewer.view.util.enums import LogType
from sohstationviewer.database.process_db import execute_db

from PySide2 import QtWidgets


class WrongDataTypeError(Exception):
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class ThreadStopped(Exception):
    """
    An exception that is raised when the user requests for the data loader
    thread to be stopped.
    """
    def __init__(self, *args, **kwargs):
        self.args = (args, kwargs)


class DataTypeModel():
    def __init__(self, data_type, tracking_box: QtWidgets.QTextBrowser,
                 folder: str, list_of_rt130_paths: List[Path] = [],
                 req_wf_chans: Union[List[str], List[int]] = [],
                 req_soh_chans: List[str] = [],
                 read_start: Optional[float] = None,
                 read_end: Optional[float] = None,
                 include_mp123zne: bool = False,
                 include_mp456uvw: bool = False,
                 creator_thread: Optional[QtCore.QThread] = None,
                 notification_signal: Optional[QtCore.Signal] = None,
                 pause_signal: Optional[QtCore.Signal] = None,
                 *args, **kwargs):
        """
        Super class for different data type to process data from data files

        :param data_type: type of the object
        :param tracking_box: widget to display tracking info
        :param folder: path to the folder of data
        :param list_of_rt130_paths: path to the folders of RT130 data
        :param req_wf_chans: requested waveform channel list
        :param req_soh_chans: requested SOH channel list
        :param read_start: requested start time to read
        :param read_end: requested end time to read
        :param include_mp123zne: if mass position channels 1,2,3 are requested
        :param include_mp456uvw: if mass position channels 4,5,6 are requested
        :param creator_thread: the thread the current DataTypeModel instance is
            being created in. If None, the DataTypeModel instance is being
            created in the main thread
        :param notification_signal: signal used to send notifications to the
            main thread. Only not None when creator_thread is not None
        :param pause_signal: signal used to notify the main thread that the
            data loader is paused.
        """
        self.data_type = data_type
        self.tracking_box = tracking_box
        self.dir = folder
        self.list_of_rt130_paths = list_of_rt130_paths
        self.req_soh_chans = req_soh_chans
        self.req_wf_chans = req_wf_chans
        self.read_start = read_start
        self.read_end = read_end
        self.include_mp123zne = include_mp123zne
        self.include_mp456uvw = include_mp456uvw
        if creator_thread is None:
            err_msg = (
                'A signal is not None while running in main thread'
            )
            assert notification_signal is None, err_msg
            assert pause_signal is None, err_msg
            self.creator_thread = QtCore.QThread()
        else:
            self.creator_thread = creator_thread
        self.notification_signal = notification_signal
        self.pause_signal = pause_signal
        """
        processing_log: [(message, type)] - record the progress of processing
        """
        self.processing_log: List[Tuple[str, LogType]] = []

        DataKey = Union[Tuple[str, str], str]

        """
        Log data: info from log channels, soh messages, text file in dict:
        {'TEXT': [str,], key:{chan_id: [str,],},}
        In which 'TEXT': is the chan_id given by sohview for text only file.
        Note: log_data for RT130's dataset has only one channel: SOH
        """
        self.log_data = {'TEXT': []}  # noqa

        """
        waveform_data: data of waveform in dict:
        {set_key: {
            'files_info': {
                chan_id - str: [{
                    'path2file': absolute path to file - str,
                    'file_name': name of file - str,
                    'sta_ids': list of station names in file - list of str
                    'chan_ids': list of channel names in file - list of str
                    'start_epoch': start time of file - float
                    'end_epoch': end time of file - float
                    'traces_total': length of data - int
                    'spr': sample rate - float
                    'read': file has been read or not - bool
                    }]
            }
            'readData': {
                chan_id - str: {
                    'samplerate': Sample rate of the data - float
                    'tracesInfo': [{
                        'samplerate': Sample rate of the data - float
                        'start_tm_epoch': Start epoch time of the trace - float
                        'end_tm_epoch': End epoch time of the trace - float
                        'size': Size of data - int
                        'times_f': np.memmap's file path for real epoch time
                            - str
                        'data_f': np.memmap's file path for data - str
                        },]
                    'times': times that has been trimmed and down- sampled for
                        plotting - np.array of float,
                    'data': data that has been trimmed and down-sampled for
                        plotting - np.array of float/int,
                    'chan_db_info': the plotting parameters got from database
                        for this channel - dict,
                    ax_wf: axes to draw the channel in WaveformWidget, to
                        combine 2 channels in need - matplotlib.axes.Axes
                }
            }
        }
        (np.memmap is used instead of data for waveform because waveform data
        are too large and make system crashed when try to merge and process)
        """
        self.waveform_data = {}

        """
        soh_data: data for SOH info in dict:
        {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    'logIdx: soh message line indexes - np.array of int,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
                logIdx: soh message line indexes that has been trimmed and
                    down-sampled for plotting - np.array of int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.soh_data = {}

        """
        mass_pos_data: data for mass position info in the format similar to
        soh_data but there are 2 axes for drawing a mass possition channel,
        one is 'ax' to display in PlottingWidget, the other is 'ax_wf' to
        display in WaveformWidget.
        Channel name for mseed mass position: VP*
        Channel name for reftek mass position: MP* (given by SOHView)
                {set_key - str or (str, str): {
            chan_id - str: {
                'org_trace': {
                    'chan_id': channel name - str
                    'samplerate': sample rate of the data - float
                    'start_tm_epoch': start epoch time of the trace - float,
                    'end_tm_epoch': end epoch time of the trace - float,
                    'times': data's real time in epoch - np.array of float,
                    'data': data - np.array of float,
                    }
                times: times that has been trimmed and down- sampled for
                    plotting - np.array of float,
                data: data that has been trimmed and down-sampled for plotting
                    - np.array of float/int,
               'chan_db_info': the plotting parameters got from database
                    for this channel - dict,
                ax: axes to draw the channel in PlottingWidget
                    - matplotlib.axes.Axes,
                ax_wf: axes to draw the channel in WaveformWidget
                    - matplotlib.axes.Axes
                }
            }
        }
        """
        self.mass_pos_data = {}

        """
        data_time: time range of data sets:
        {setKey: [earliestepoch, latestepoch]} - {str: [float, float],}
        """
        self.data_time: Dict[DataKey, List[float]] = {}

        """
        The given data may include more than one data set which is station_id
        in mseed or (unit_id, exp_no) in reftek. User are allow to choose which
        data set to be displayed
        selected_key: str - key of the data set to be displayed
        """
        self.selected_key: Optional[str] = None

        """
        gaps: gaps info in dict:
        {set_key: [list of gaps]} - {str: [[float, float],],}
        """
        self.gaps: Dict[DataKey, List[List[float]]] = {}

        """
         tmp_dir: str - dir to keep memmap files.
         Will be deleted when object is deleted
        """
        self.tmp_dir_obj: TemporaryDirectory = TemporaryDirectory()
        self.tmp_dir = self.tmp_dir_obj.name
        self.save_temp_data_folder_to_database()

        self._pauser = QtCore.QSemaphore()
        self.pause_response = None

        self.gps_points: List[GPSPoint] = []

    def __del__(self):
        print("delete dataType Object")
        try:
            del self.tmp_dir_obj
        except OSError as e:
            self.track_info(
                "Error deleting %s : %s" % (self.tmp_dir, e.strerror),
                LogType.ERROR)
            print("Error deleting %s : %s" % (self.tmp_dir, e.strerror))
        print("finish deleting")

    def has_data(self) -> bool:
        """
        :return bool - True if there is any data can be read.
                       False if there is no valid data
        """
        if (len(self.log_data) == 1 and len(self.soh_data) == 0 and
                len(self.mass_pos_data) == 0 and len(self.waveform_data) == 0):
            return False
        return True

    def track_info(self, text: str, type: LogType) -> None:
        """
        Display tracking info in tracking_box.
        Add all errors/warnings to processing_log.
        :param text: str - message to display
        :param type: str - type of message (error/warning/info)
        """
        # display_tracking_info updates a QtWidget, which can only be done in
        # the main thread. So, if we are running in a background thread
        # (i.e. self.creator_thread is not None), we need to use signal slot
        # mechanism to ensure that display_tracking_info is run in the main
        # thread.
        if self.notification_signal is None:
            display_tracking_info(self.tracking_box, text, type)
        else:
            self.notification_signal.emit(self.tracking_box, text, type)
        if type != LogType.INFO:
            self.processing_log.append((text, type))

    @classmethod
    def create_data_object(cls, data_type, tracking_box, folder,
                           list_of_rt130_paths,
                           req_wf_chans=[], req_soh_chans=[],
                           read_start=0, read_end=constants.HIGHEST_INT,
                           include_mp123=False, include_mp456=False):
        """
        Create a DataTypeModel object, with the concrete class being based on
        data_type. Run on the same thread as its caller, and so will block the
        GUI if called on the main thread. Do not call this method directly.
        Instead, call the wrapper controller.processing.load_data.

        :param data_type: str - type of data read
        :param tracking_box: QTextBrowser - widget to display tracking info
        :param folder: [str,] - the data directory
        :param list_of_rt130_paths: [Path,]: list of RT130 directories
        :param req_wf_chans: [str,] - requested waveform channel list
        :param req_soh_chans: [str,] - requested soh channel list
        :param read_start: [float,] - start time of read data
        :param read_end: [float,] - finish time of read data
        :param include_mp123: bool - if mass pos 1,2,3 are requested
        :param include_mp456: bool - if mass pos 4,5,6 are requested
        :return: DataTypeModel - object that keep the data read from
            folder
        """
        if data_type == 'RT130':
            from sohstationviewer.model.reftek.reftek import RT130
            data_object = RT130(
                data_type, tracking_box, folder, list_of_rt130_paths,
                reqWFChans=req_wf_chans, reqSOHChans=req_soh_chans,
                readStart=read_start, readEnd=read_end,
                include_mp123=include_mp123, include_mp456=include_mp456)
        else:
            from sohstationviewer.model.mseed.mseed import MSeed
            data_object = MSeed(
                data_type, tracking_box, folder, list_of_rt130_paths,
                reqWFChans=req_wf_chans, reqSOHChans=req_soh_chans,
                readStart=read_start, readEnd=read_end,
                include_mp123=include_mp123, include_mp456=include_mp456)
        return data_object

    def pause(self) -> None:
        """
        Pause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Note: due to how this is implemented, each call to pause will require
        a corresponding call to unpause. Thus, it is inadvisable to call this
        method more than once.

        Caution: not safe to call in the main thread. Unless a background
        thread releases the semaphore, the whole program will freeze.
        """
        self._pauser.acquire()

    @QtCore.Slot()
    def unpause(self):
        """
        Unpause the thread this DataTypeModel instance is in. Works by trying
        to acquire a semaphore that is not available, which causes the thread
        to block.

        Caution: due to how this is implemented, if unpause is called before
        pause, the thread will not be paused until another call to pause is
        made. Also, like pause, each call to unpause must be matched by another
        call to pause for everything to work.
        """
        self._pauser.release()

    @QtCore.Slot()
    def receive_pause_response(self, response: object):
        """
        Receive a response to a request made to another thread and unpause the
        calling thread.

        :param response: the response to the request made
        :type response: object
        """
        self.pause_response = response
        self.unpause()

    @classmethod
    def get_empty_instance(cls) -> DataTypeModel:
        """
        Create an empty data object. Useful if a DataTypeModel instance is
        needed, but it is undesirable to load a data set. Basically wraps
        __new__().

        :return: an empty data object
        :rtype: DataTypeModel
        """
        return cls.__new__(cls)

    def save_temp_data_folder_to_database(self):
        execute_db(f'UPDATE PersistentData SET FieldValue="{self.tmp_dir}" '
                   f'WHERE FieldName="tempDataDirectory"')

    def check_not_found_soh_chans(self):
        not_found_soh_chans = [
            c for c in self.req_soh_chans
            if c not in self.soh_data[self.selected_key].keys()]
        if not_found_soh_chans != []:
            msg = (f"No data found for soh channels: "
                   f"{', '.join( not_found_soh_chans)}")
            self.processing_log.append((msg, LogType.WARNING))
